
// 12/11 Removed WEP flag from ic_caps in if_ath.c 
// 12/11 Added flag to ic_caps in if_ath.c to allow txpower via iwconfig
// 12/10 Added wlansniff header (but w/o 802.11 hdr)
#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/random.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/cache.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include <linux/if_arp.h>

#include <asm/uaccess.h>

#include "gt.h"

#define	AR_DEBUG
#include "if_athvar.h"

#ifdef GT_DEV

// Expand SKB's headroom and tailroom by HEAD_NEED and TAIL_NEED amount. 
// Return the modified skb upon success and NULL, otherwise. 
static struct sk_buff* 
expand_skb(struct sk_buff *skb, int head_need, int tail_need) 
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
    if (head_need > skb_headroom(skb) || tail_need > skb_tailroom(skb)) {
	struct sk_buff *nskb;

	nskb = dev_alloc_skb(skb->len + head_need + tail_need);
	if (nskb == NULL) {
	    //dev_kfree_skb_any(skb);
	    return NULL;
	}

	skb_reserve(nskb, head_need);
	memcpy(skb_put(nskb, skb->len), skb->data, skb->len);
	dev_kfree_skb_any(skb);
	return nskb;
    }
    return skb; // has enough tail and head rooms

#else // Linux 2.4.0 or newer 

    head_need -= skb_headroom(skb);
    tail_need -= skb_tailroom(skb);
    
    if (head_need > 0 || tail_need > 0) {
	if (pskb_expand_head(skb, head_need > 0 ? head_need : 0,
			     tail_need > 0 ? tail_need : 0,
			     GFP_ATOMIC)) {
	    printk(KERN_DEBUG "%ld expand_skb failed to " 
		   "reallocate skb buffer\n", jiffies);
	    return NULL;
	}
	//printk(KERN_DEBUG "%ld expand_skb success! But consider per-allocating buffer!\n", jiffies);
    }
    return skb;
#endif
}

// clones skb and adds struct sniff_header to the cloned copy, then passes it
// up the stack with PACKET_OTHERHOST type
void gtif_ath_rx(struct net_device *dev, struct sk_buff *skb, struct ath_desc *ds)
{
    struct sk_buff *cskb = skb_clone(skb, GFP_ATOMIC);
    if (cskb)
	cskb = expand_skb(cskb, sizeof(struct sniff_header), 0);
    
    if (cskb) {
	struct ath_softc *sc = dev->priv;
	struct ieee80211com *ic  = &sc->sc_ic;
	const HAL_RATE_TABLE *rt = sc->sc_currates;
	//int intrate = 0;

	struct sniff_header *sniff = (struct sniff_header *) skb_push(cskb, sizeof(struct sniff_header));
	
	sniff->dev_type = SNIFF_HOSTAP;
	sniff->data[0] = 1;

	wlan_ng_prism2_header *hdr;
	hdr = (wlan_ng_prism2_header *)(&sniff->data[1]);

#define LWNG_SETVAL(f,m,j,k,q) \
hdr->f.did = LWNG_CAP_DID_BASE | (m << 12); \
hdr->f.status = j; hdr->f.len = k; hdr->f.data = q;
        LWNG_SETVAL(hosttime, 1, 0, 4, jiffies);
	LWNG_SETVAL(mactime, 2, 0, 4, ath_hal_gettsf32(sc->sc_ah));
	LWNG_SETVAL(channel, 3, 0, 4, 
		    ieee80211_mhz2ieee(ic->ic_ibss_chan->ic_freq,0));
	//LWNG_SETVAL(rssi, 4, 0, 4, ds->ds_rxstat.rs_rssi);
	LWNG_SETVAL(rssi, 4, P80211ENUM_msgitem_status_no_value, 4, 0);
	LWNG_SETVAL(sq, 5, 1 /* no value */, 4, 0);
	LWNG_SETVAL(signal, 6, 0, 4, ds->ds_rxstat.rs_rssi);
	//LWNG_SETVAL(signal, 6, P80211ENUM_msgitem_status_no_value, 4, 0);
	LWNG_SETVAL(noise, 7, P80211ENUM_msgitem_status_no_value, 4, 0);

	/*
	intrate = 		    (rt != NULL ?
	(rt->info[rt->rateCodeToIndex[ds->ds_rxstat.rs_rate]].dot11Rate & IEEE80211_RATE_VAL) : 2);
	printk("recv packet rate=%d\n", intrate);
	LWNG_SETVAL(rate, 8, 0, 4, intrate);
	*/

	LWNG_SETVAL(rate, 8, 0, 4, (rt != NULL ?
	(rt->info[rt->rateCodeToIndex[ds->ds_rxstat.rs_rate]].dot11Rate & IEEE80211_RATE_VAL) : 2));
	LWNG_SETVAL(istx, 9, 0, 4, P80211ENUM_truth_false);
	LWNG_SETVAL(frmlen, 10, 0, 4, ds->ds_rxstat.rs_datalen);
	

	cskb->mac.raw = cskb->data; //@GT: raw socket will get the data beginning with sniff header
	//skb_pull(cskb, sizeof(struct sniff_header)); // ?? why do we need this?
	cskb->pkt_type = PACKET_OTHERHOST; // only raw socket view will see
	cskb->protocol = htons(ETH_P_IRDA); // for sniffing programs to distinguish between appl level pkt and sniff pkt
	netif_rx(cskb);
    }
}
#endif  // GT_DEV
